#!/usr/bin/env python
import setpath
import unittest
import os

from bike import testdata
from bike.query.findDefinition import findModule, findPackage,\
     findAllPossibleDefinitionsByCoords
from bike.query.getTypeOf import getTypeOf
from bike.testutils import trimLines, createSourceNodeAt
from bike.parsing.extended_ast import Root

        
class TestFindDefinitionByCoords(unittest.TestCase):
    def test_findsClassRef(self):
        src=trimLines("""
        class TheClass:
            pass
        a = TheClass()
        """)
        root = createSourceNodeAt(Root(),src,"mymodule")
        module = getTypeOf(root,"mymodule")
        defn = [x for x in findAllPossibleDefinitionsByCoords(module,3,6)]
        assert defn[0].filename == os.path.abspath("mymodule.py")
        assert defn[0].lineno == 1
        assert defn[0].colno == 6
        assert defn[0].confidence == 100

    def tests_findsMethodRef(self):
        src=trimLines("""
        class TheClass:
            def theMethod(self):
                pass
        a = TheClass()
        a.theMethod()
        """)
        root = createSourceNodeAt(Root(),src,"mymodule")
        module = getTypeOf(root,"mymodule")
        defn = [x for x in findAllPossibleDefinitionsByCoords(module,5,3)]
        assert defn[0].filename == os.path.abspath("mymodule.py")
        assert defn[0].lineno == 2
        assert defn[0].colno == 8
        assert defn[0].confidence == 100
        

    def test_returnsOtherMethodsWithSameName(self):
        src=trimLines("""
        class TheClass:
            def theMethod(self):
                pass
        a = SomeOtherClass()
        a.theMethod()
        """)
        root = createSourceNodeAt(Root(),src,"mymodule")
        module = getTypeOf(root,"mymodule")
        defn = [x for x in findAllPossibleDefinitionsByCoords(module,5,3)]
        assert defn[0].filename == os.path.abspath("mymodule.py")
        assert defn[0].lineno == 2
        assert defn[0].colno == 8
        assert defn[0].confidence == 50




    def test_findsTemporaryDefinition(self):
        src=trimLines("""
        a = 3
        b = a + 1
        """)
        root = createSourceNodeAt(Root(),src,"mymodule")
        module = getTypeOf(root,"mymodule")
        defn = [x for x in findAllPossibleDefinitionsByCoords(module,2,4)]
        assert defn[0].filename == os.path.abspath("mymodule.py")
        assert defn[0].lineno == 1
        assert defn[0].colno == 0
        assert defn[0].confidence == 100

    def test_findsArgumentDefinition(self):
        src=trimLines("""
        def someFunction(a):
            b = a + 1
        """)
        root = createSourceNodeAt(Root(),src,"mymodule")
        module = getTypeOf(root,"mymodule")
        defn = [x for x in findAllPossibleDefinitionsByCoords(module,2,8)]
        assert defn[0].filename == os.path.abspath("mymodule.py")
        assert defn[0].lineno == 1
        assert defn[0].colno == 17
        assert defn[0].confidence == 100

    def test_findsClassInstanceDefinition(self):
        src=trimLines("""
        class TheClass():
            pass
        a = TheClass()
        print a
        """)
        root = createSourceNodeAt(Root(),src,"mymodule")
        module = getTypeOf(root,"mymodule")
        defn = [x for x in findAllPossibleDefinitionsByCoords(module,4,6)]
        assert defn[0].filename == os.path.abspath("mymodule.py")
        assert defn[0].lineno == 3
        assert defn[0].colno == 0
        assert defn[0].confidence == 100

    def test_findsDefinitionInParentScope(self):
        src=trimLines("""
        a = 3
        def foo(self):
            b = a + 1
        """)
        root = createSourceNodeAt(Root(),src,"mymodule")
        module = getTypeOf(root,"mymodule")
        defn = [x for x in findAllPossibleDefinitionsByCoords(module,3,8)]
        assert defn[0].filename == os.path.abspath("mymodule.py")
        assert defn[0].lineno == 1
        assert defn[0].colno == 0
        assert defn[0].confidence == 100

    def test_findsDefinitionWithinFunction(self):
        src=trimLines("""
        def foo(yadda):
            a = someFunction()
            print a
        """)
        root = createSourceNodeAt(Root(),src,"mymodule")
        module = getTypeOf(root,"mymodule")
        defn = [x for x in findAllPossibleDefinitionsByCoords(module,3,10)]
        assert defn[0].filename == os.path.abspath("mymodule.py")
        assert defn[0].lineno == 2
        assert defn[0].colno == 4
        assert defn[0].confidence == 100
        

    def test_findsDefinitionFromSubsequentAssignment(self):
        src=trimLines("""
        def foo(yadda):
            a = 3
            print a
            a = 5
        """)
        root = createSourceNodeAt(Root(),src,"mymodule")
        module = getTypeOf(root,"mymodule")
        defn = [x for x in findAllPossibleDefinitionsByCoords(module,4,4)]
        assert defn[0].filename == os.path.abspath("mymodule.py")
        assert defn[0].lineno == 2
        assert defn[0].colno == 4
        assert defn[0].confidence == 100

    def test_findsDefinitionFromDefinition(self):
        src=trimLines("""
        def foo(yadda):
            a = 3
            print a
            a = 5
        """)
        root = createSourceNodeAt(Root(),src,"mymodule")
        module = getTypeOf(root,"mymodule")
        defn = [x for x in findAllPossibleDefinitionsByCoords(module,2,4)]
        assert defn[0].filename == os.path.abspath("mymodule.py")
        assert defn[0].lineno == 2
        assert defn[0].colno == 4
        assert defn[0].confidence == 100


    def test_findsClassRefUsingFromImportStatement(self):
        src=trimLines("""
        from a.b.bah import TheClass
        """)
        classsrc=trimLines("""
        class TheClass:
            pass
        """)
        root = createSourceNodeAt(Root(),src,"a.foo")
        root = createSourceNodeAt(root, classsrc, "a.b.bah")
        module = getTypeOf(root,"a.foo")
        defn = [x for x in findAllPossibleDefinitionsByCoords(module,1,21)]
        assert defn[0].filename == os.path.abspath(os.path.join("a","b","bah.py"))
        assert defn[0].lineno == 1
        assert defn[0].colno == 6
        assert defn[0].confidence == 100


    def test_findsVariableRefUsingFromImportStatement(self):
        importsrc=trimLines("""
        from a.b.bah import mytext
        print mytext
        """)
        src=trimLines("""
        mytext = 'hello'
        """)
        root = createSourceNodeAt(Root(),importsrc,"a.foo")
        root = createSourceNodeAt(root, src, "a.b.bah")
        module = getTypeOf(root,"a.foo")
        defn = [x for x in findAllPossibleDefinitionsByCoords(module,2,6)]
        assert defn[0].filename == os.path.abspath(os.path.join("a","b","bah.py"))
        assert defn[0].lineno == 1
        assert defn[0].colno == 0
        assert defn[0].confidence == 100


    def test_findsVariableRefUsingImportStatement(self):
        importsrc=trimLines("""
        import a.b.bah
        print a.b.bah.mytext
        """)
        src=trimLines("""
        mytext = 'hello'
        """)
        root = createSourceNodeAt(Root(),importsrc,"a.foo")
        root = createSourceNodeAt(root, src, "a.b.bah")
        module = getTypeOf(root,"a.foo")
        defn = [x for x in findAllPossibleDefinitionsByCoords(module,2,14)]
        assert defn[0].filename == os.path.abspath(os.path.join("a","b","bah.py"))
        assert defn[0].lineno == 1
        assert defn[0].colno == 0
        assert defn[0].confidence == 100


    def test_findsVariableRefUsingFromImportStarStatement(self):
        importsrc=trimLines("""
        from a.b.bah import *
        print mytext
        """)
        src=trimLines("""
        mytext = 'hello'
        """)
        root = createSourceNodeAt(Root(),importsrc,"a.foo")
        root = createSourceNodeAt(root, src, "a.b.bah")
        module = getTypeOf(root,"a.foo")
        defn = [x for x in findAllPossibleDefinitionsByCoords(module,2,6)]
        assert defn[0].filename == os.path.abspath(os.path.join("a","b","bah.py"))
        assert defn[0].lineno == 1
        assert defn[0].colno == 0
        assert defn[0].confidence == 100

    def test_findsVariableRefUsingFromPackageImportModuleStatement(self):
        importsrc=trimLines("""
        from a.b import bah
        print bah.mytext
        """)
        src=trimLines("""
        mytext = 'hello'
        """)
        root = createSourceNodeAt(Root(),importsrc,"a.b.foo")
        root = createSourceNodeAt(root, src, "a.b.bah")
        module = getTypeOf(root,"a.b.foo")
        defn = [x for x in findAllPossibleDefinitionsByCoords(module,2,10)]
        assert defn[0].filename == os.path.abspath(os.path.join("a","b","bah.py"))
        assert defn[0].lineno == 1
        assert defn[0].colno == 0
        assert defn[0].confidence == 100

    def test_findsImportedVariableRefInAFunctionArg(self):
        importsrc=trimLines("""
        from a.b import bah
        someFunction(bah.mytext)
        """)
        src=trimLines("""
        mytext = 'hello'
        """)
        root = createSourceNodeAt(Root(),importsrc,"a.b.foo")
        root = createSourceNodeAt(root, src, "a.b.bah")
        module = getTypeOf(root,"a.b.foo")
        defn = [x for x in findAllPossibleDefinitionsByCoords(module,2,17)]
        assert defn[0].filename == os.path.abspath(os.path.join("a","b","bah.py"))
        assert defn[0].lineno == 1
        assert defn[0].colno == 0
        assert defn[0].confidence == 100


    def test_findsVariableRefUsingFromImportStatementInFunction(self):
        importsrc=trimLines("""
        def foo:
            from a.b.bah import mytext
            print mytext
        """)
        src=trimLines("""
        mytext = 'hello'
        """)
        root = createSourceNodeAt(Root(),importsrc,"a.foo")
        root = createSourceNodeAt(root, src, "a.b.bah")
        module = getTypeOf(root,"a.foo")
        defn = [x for x in findAllPossibleDefinitionsByCoords(module,3,10)]
        assert defn[0].filename == os.path.abspath(os.path.join("a","b","bah.py"))
        assert defn[0].lineno == 1
        assert defn[0].colno == 0
        assert defn[0].confidence == 100

    def test_findsVariableRefByImportingModule(self):
        importsrc=trimLines("""
        import a.b.bah
        print a.b.bah.mytext
        """)
        src=trimLines("""
        mytext = 'hello'
        """)
        root = createSourceNodeAt(Root(),importsrc,"a.b.foo")
        root = createSourceNodeAt(root, src, "a.b.bah")
        module = getTypeOf(root,"a.b.foo")
        defn = [x for x in findAllPossibleDefinitionsByCoords(module,2,14)]
        assert defn[0].filename == os.path.abspath(os.path.join("a","b","bah.py"))
        assert defn[0].lineno == 1
        assert defn[0].colno == 0
        assert defn[0].confidence == 100

    def test_doesntfindVariableRefOfUnimportedModule(self):
        importsrc=trimLines("""
        # a.b.bah not imported
        print a.b.bah.mytext
        """)
        src=trimLines("""
        mytext = 'hello'
        """)
        root = createSourceNodeAt(Root(),importsrc,"a.b.foo")
        root = createSourceNodeAt(root, src, "a.b.bah")
        module = getTypeOf(root,"a.b.foo")
        defn = [x for x in findAllPossibleDefinitionsByCoords(module,2,14)]
        self.assertEqual(defn,[])



    def test_findsSelfAttributeDefinition(self):
        src=trimLines("""
        class MyClass:
           def __init__(self):
               self.a = 'hello'
           def myMethod(self):
               print self.a
        """)
        root = createSourceNodeAt(Root(),src,"mymodule")
        module = getTypeOf(root,"mymodule")
        defn = [x for x in findAllPossibleDefinitionsByCoords(module,5,18)]
        assert defn[0].filename == os.path.abspath("mymodule.py")
        assert defn[0].lineno == 3
        assert defn[0].colno == 12
        assert defn[0].confidence == 100

    def test_findsSelfAttributeDefinitionFromSamePlace(self):
        src=trimLines("""
        class MyClass:
           def __init__(self):
               self.a = 'hello'
           def myMethod(self):
               print self.a
        """)
        root = createSourceNodeAt(Root(),src,"mymodule")
        module = getTypeOf(root,"mymodule")
        defn = [x for x in findAllPossibleDefinitionsByCoords(module,3,12)]
        assert defn[0].filename == os.path.abspath("mymodule.py")
        assert defn[0].lineno == 3
        assert defn[0].colno == 12
        assert defn[0].confidence == 100


    def test_findsSelfAttributeDefinition(self):
        src=trimLines("""
        class MyClass:
            def someOtherFn(self):
                pass
            def load(self, source):
                # fastparser ast
                self.fastparseroot = fastparser(source,self.modulename)
        """)
        root = createSourceNodeAt(Root(),src,"mymodule")
        module = getTypeOf(root,"mymodule")
        defn = [x for x in findAllPossibleDefinitionsByCoords(module,6,14)]
        assert defn[0].filename == os.path.abspath("mymodule.py")
        assert defn[0].lineno == 6
        assert defn[0].colno == 13
        assert defn[0].confidence == 100


    def test_findsDefnOfInnerClass(self):
        src = trimLines("""
        class TheClass:
            class TheClass:
                pass
        a = TheClass.TheClass()
        """)
        root = createSourceNodeAt(Root(),src,"mymodule")
        module = getTypeOf(root,"mymodule")
        defn = [x for x in findAllPossibleDefinitionsByCoords(module,4,14)]
        assert defn[0].filename == os.path.abspath("mymodule.py")
        assert defn[0].lineno == 2
        assert defn[0].colno == 10
        assert defn[0].confidence == 100

    def test_findsDefnOfOuterClass(self):
        src = trimLines("""
        class TheClass:
            class TheClass:
                pass
        a = TheClass.TheClass()
        """)
        root = createSourceNodeAt(Root(),src,"mymodule")
        module = getTypeOf(root,"mymodule")
        defn = [x for x in findAllPossibleDefinitionsByCoords(module,4,4)]
        assert defn[0].filename == os.path.abspath("mymodule.py")
        assert defn[0].lineno == 1
        assert defn[0].colno == 6
        assert defn[0].confidence == 100


    def test_findsClassDeclaredIn__init__Module(self):
        importsrc=trimLines("""
        class TheClass:
            pass
        """)
        src=trimLines("""
        from a import TheClass
        c = TheClass()
        """)
        root = createSourceNodeAt(Root(),importsrc,"a.__init__")
        root = createSourceNodeAt(root, src, "mymodule")
        module = getTypeOf(root,"mymodule")
        defn = [x for x in findAllPossibleDefinitionsByCoords(module,2,6)]
        assert defn[0].filename == os.path.abspath(os.path.join("a",
                                                                "__init__.py"))
        assert defn[0].lineno == 1
        assert defn[0].colno == 6
        assert defn[0].confidence == 100



class TestFindModule(unittest.TestCase):
    def test_findsModuleFromName(self):
        root = createSourceNodeAt(Root(),"pass","bah")
        assert findModule(root,"bah").name == "bah"
        assert findModule(root,"bah").fqn == "bah"

    def test_findsModuleFromFqn(self):
        root = createSourceNodeAt(Root(),"pass","a.b.bah")
        assert findModule(root,"a.b.bah").name == "bah"
        assert findModule(root,"a.b.bah").fqn == "a.b.bah"

    def test_findsModuleFromPackageScope(self):
        root = createSourceNodeAt(Root(),"pass","a.b.bah")
        package = findPackage(root,"a.b")
        assert findModule(package,"bah").name == "bah"
        assert findModule(package,"bah").fqn == "a.b.bah"

if __name__ == "__main__":
    unittest.main()
