#!/usr/bin/env python
import setpath
import unittest
import os
from bike import testdata
from bike.testutils import trimLines, createSourceNodeAt
from bike import testdata
from bike.parsing.extended_ast import Root
from findReferences import findReferences, findReferencesIncludingDefn
from bike.query.getTypeOf import getTypeOf

class TestFindReferences(unittest.TestCase):
    def test_findsSimpleReferencesGivenAssignment(self):
        src=trimLines("""
        def foo():
            a = 3
            print a
        """)
        root = createSourceNodeAt(Root(),src,"mymodule")
        sourcenode = root.getSource("mymodule")
        refs = [x for x in findReferences(sourcenode,2,4)]
        assert refs[0].filename == os.path.abspath("mymodule.py")
        assert refs[0].lineno == 3
        assert refs[0].colno == 10
        assert refs[0].confidence == 100

    def test_findsSimpleReferencesGivenReference(self):
        src=trimLines("""
        def foo():
            a = 3
            print a
        """)
        root = createSourceNodeAt(Root(),src,"mymodule")
        sourcenode = root.getSource("mymodule")
        refs = [x for x in findReferencesIncludingDefn(sourcenode,3,10)]
        assert refs[0].filename == os.path.abspath("mymodule.py")
        assert refs[0].lineno == 2
        assert refs[0].colno == 4
        assert refs[0].confidence == 100


    def test_findsReferencesToOtherAssignments(self):
        src=trimLines("""
        def foo():
            a = 3
            a = 4
        """)
        root = createSourceNodeAt(Root(),src,"mymodule")
        sourcenode = root.getSource("mymodule")
        refs = [x for x in findReferences(sourcenode,2,4)]
        assert refs[0].filename == os.path.abspath("mymodule.py")
        assert refs[0].lineno == 3
        assert refs[0].colno == 4
        assert refs[0].confidence == 100

    def test_findsFunctionArg(self):
        src=trimLines("""
        def foo(a):
            print a
        """)
        root = createSourceNodeAt(Root(),src,"mymodule")
        sourcenode = root.getSource("mymodule")
        refs = [x for x in findReferencesIncludingDefn(sourcenode,2,10)]
        assert refs[0].filename == os.path.abspath("mymodule.py")
        assert refs[0].lineno == 1
        assert refs[0].colno == 8
        assert refs[0].confidence == 100

    def test_findsFunctionArgWithDefault(self):
        src=trimLines("""
        def foo(a=None, b=None):
            print a, b
        """)
        root = createSourceNodeAt(Root(),src,"mymodule")
        sourcenode = root.getSource("mymodule")
        refs = [x for x in findReferences(sourcenode,1,4)]
        self.assertEquals(refs, [])
        refs = [x for x in findReferencesIncludingDefn(sourcenode,2,13)]
        assert refs[0].filename == os.path.abspath("mymodule.py")
        assert refs[0].lineno == 1
        assert refs[0].colno == 16
        assert refs[0].confidence == 100

    def test_findsReferencesGivenFunctionArg(self):
        src=trimLines("""
        def foo(a):
            print a
        """)
        root = createSourceNodeAt(Root(),src,"mymodule")
        sourcenode = root.getSource("mymodule")
        refs = [x for x in findReferences(sourcenode,1,8)]
        assert refs[0].filename == os.path.abspath("mymodule.py")
        assert refs[0].lineno == 2
        assert refs[0].colno == 10
        assert refs[0].confidence == 100


    def test_findsVariableRefInImportStatementUsingFromImportStatement(self):
        importsrc=trimLines("""
        from a.b.bah import mytext
        """)
        src=trimLines("""
        mytext = 'hello'
        """)
        root = createSourceNodeAt(Root(),importsrc,"a.foo")
        root = createSourceNodeAt(root, src, "a.b.bah")
        module = getTypeOf(root,"a.b.bah")
        refs = [x for x in findReferences(module.getSourceNode(),1,1)]
        assert refs[0].filename == os.path.abspath(os.path.join("a","foo.py"))
        assert refs[0].lineno == 1
        assert refs[0].colno == 20
        assert refs[0].confidence == 100

    def test_findsVariableRefUsingFromImportStatement(self):
        importsrc=trimLines("""
        from a.b.bah import mytext
        print mytext
        """)
        src=trimLines("""
        mytext = 'hello'
        """)
        root = createSourceNodeAt(Root(),importsrc,"a.foo")
        root = createSourceNodeAt(root, src, "a.b.bah")
        module = getTypeOf(root,"a.b.bah")
        refs = [x for x in findReferences(module.getSourceNode(),1,1)]
        #print refs
        assert refs[1].filename == os.path.abspath(os.path.join("a","foo.py"))
        assert refs[1].lineno == 2
        assert refs[1].colno == 6
        assert refs[1].confidence == 100

    def test_findsImportedVariableRefInAFunctionArg(self):
        importsrc=trimLines("""
        from a.b import bah
        someFunction(bah.mytext)
        """)
        src=trimLines("""
        mytext = 'hello'
        """)
        root = createSourceNodeAt(Root(),importsrc,"a.foo")
        root = createSourceNodeAt(root, src, "a.b.bah")
        module = getTypeOf(root,"a.b.bah")
        refs = [x for x in findReferences(module.getSourceNode(),1,1)]
        assert refs[0].filename == os.path.abspath(os.path.join("a","foo.py"))
        assert refs[0].lineno == 2
        assert refs[0].colno == 17
        assert refs[0].confidence == 100

    def test_getsReferenceOfSimpleMethodCall(self):
        src = trimLines("""
        from b.bah import TheClass
        a = TheClass()
        a.theMethod()
        """)
        root = createSourceNodeAt(Root(),src,"a.foo")
        root = createSourceNodeAt(root, testdata.TheClass, "a.b.bah")
        module = getTypeOf(root,"a.foo")
        refs = [x for x in findReferences(module.getSourceNode(),3,2)]
        self.assertEqual(refs[0].filename,
                         os.path.abspath(os.path.join("a","foo.py")))
        self.assertEqual(refs[0].modulefqn,"a.foo")
        self.assertEqual(refs[0].lineno,3)
        self.assertEqual(refs[0].colno,2)
        self.assertEqual(refs[0].colno,2)

    def test_findsRefToSelfAttribute(self):
        src=trimLines("""
        class MyClass:
           def __init__(self):
               self.a = 'hello'
           def myMethod(self):
               print self.a
        """)
        root = createSourceNodeAt(Root(),src,"mymodule")
        module = getTypeOf(root,"mymodule")
        refs = [x for x in findReferences(module.getSourceNode(),3,12)]
        assert refs[0].filename == os.path.abspath("mymodule.py")
        assert refs[0].lineno == 5
        assert refs[0].colno == 18
        assert refs[0].confidence == 100


class FindReferencesToMethod(unittest.TestCase):
    def test_findsReferenceOfSimpleMethodCall(self):
        src = trimLines("""
        from b.bah import TheClass
        a = TheClass()
        a.theMethod()
        """)
        root = createSourceNodeAt(Root(),src,"a.foo")
        root = createSourceNodeAt(root, testClass, "a.b.bah")
        module = getTypeOf(root,"a.b.bah")
        refs = [x for x in findReferences(module.getSourceNode(),2,8)
                if x.confidence == 100]
        self.assertEqual(refs[0].filename,
                         os.path.abspath(os.path.join("a","foo.py")))
        self.assertEqual(refs[0].lineno,3)
        self.assertEqual(refs[0].colno,2)
        self.assertEqual(refs[0].colno,2)

    def test_getsReferenceOfMethodCallFromClassImportedWithAlias(self):
        src = trimLines("""
        import b.bah.TheClass as MyTheClass
        
        def foo():
            a = MyTheClass()
            a.theMethod()
        """)
        root = createSourceNodeAt(Root(),src,"a.foo")
        root = createSourceNodeAt(root, testClass, "a.b.bah")
        module = getTypeOf(root,"a.b.bah")
        refs = [x for x in findReferences(module.getSourceNode(),2,8)
                if x.confidence == 100]
        self.assertEqual(refs[0].filename,
                         os.path.abspath(os.path.join("a","foo.py")))
        self.assertEqual(refs[0].modulefqn,"a.foo")
        self.assertEqual(refs[0].lineno,5)
        self.assertEqual(refs[0].colno,6)


    def test_getsReferenceOfMethodCallWhenInstanceReturnedByFunction(self):
        src = trimLines("""
        from b.bah import TheClass
        
        def foo():
            return TheClass()
        a = foo()
        a.theMethod()
        """)
        root = createSourceNodeAt(Root(),src,"a.foo")
        root = createSourceNodeAt(root, testClass, "a.b.bah")
        module = getTypeOf(root,"a.b.bah")
        refs = [x for x in findReferences(module.getSourceNode(),2,8)
                if x.confidence == 100]
        self.assertEqual(refs[0].filename,
                         os.path.abspath(os.path.join("a","foo.py")))
        self.assertEqual(refs[0].modulefqn,"a.foo")
        self.assertEqual(refs[0].lineno,6)
        self.assertEqual(refs[0].colno,2)

    def test_getsReferenceOfMethodCallInSameClass(self):
        src = trimLines("""
        class TheClass:
            def theMethod(self):
                pass
            def anotherMethod(self):
                self.theMethod()
        """)
        root = createSourceNodeAt(Root(),src,"a.foo")
        module = getTypeOf(root,"a.foo")
        refs = [x for x in findReferences(module.getSourceNode(),2,8)
                if x.confidence == 100]
        self.assertEqual(refs[0].filename,
                         os.path.abspath(os.path.join("a","foo.py")))
        self.assertEqual(refs[0].modulefqn,"a.foo")
        self.assertEqual(refs[0].lineno,5)
        self.assertEqual(refs[0].colno,13)

    def test_getsReferenceOfMethodOnBaseClassInstance(self):
        src = trimLines("""
        class root:
            def theMethod():
                pass
                
        class a(root):
            def theMethod():
                pass
                
        class b(root):
            pass
            
        class TheClass(b):
            def theMethod(self):
                pass
                
        rootinstance = root()
        rootinstance.theMethod()
        """)
        root = createSourceNodeAt(Root(),src,"a.foo")

        module = getTypeOf(root,"a.foo")
        refs = [x for x in findReferences(module.getSourceNode(),13,8)
                if x.confidence == 100]
        self.assertEqual(refs[2].filename,
                         os.path.abspath(os.path.join("a","foo.py")))
        self.assertEqual(refs[2].modulefqn,"a.foo")
        self.assertEqual(refs[2].lineno,17)
        self.assertEqual(refs[2].colno,13)
        
    def test_doesntGetReferenceToMethodWhenObjectCreatedInChildScopeToMethodReference(self):
        src = trimLines("""
        from b.bah import TheClass
        a = AnotherClass()
        def foo():
            a = TheClass()
        a.theMethod()
        """)
        root = createSourceNodeAt(Root(),src,"a.foo")
        root = createSourceNodeAt(root, testClass, "a.b.bah")
        module = getTypeOf(root,"a.b.bah")
        refs = [x for x in findReferences(module.getSourceNode(),2,8)
                if x.confidence == 100]
        assert refs == []
        
    def test_renamesMethodReferenceOfInstanceCreatedInSubsequentFunction(self):
        src = trimLines("""
        class TheClass:
            def theMethod():
                pass
        class NotTheClass:
            def theMethod():
                pass
            
        def foo():
            a = bah()
            a.theMethod()
            
        def bah():
            return TheClass()
        """)
        root = createSourceNodeAt(Root(),src,"a.foo")
        module = getTypeOf(root,"a.foo")
        refs = [x for x in findReferences(module.getSourceNode(),2,8)
                if x.confidence == 100]
        self.assertEqual(refs[0].filename,
                         os.path.abspath(os.path.join("a","foo.py")))
        self.assertEqual(refs[0].modulefqn,"a.foo")
        self.assertEqual(refs[0].lineno,10)
        self.assertEqual(refs[0].colno,6)


    def test_getsReferenceInMiddleOfBiggerCompoundCall(self):
        src = trimLines("""
        class TheClass:
            def theMethod(self): return AnotherClass()
        TheClass().theMethod().anotherMethod()
        """)
        root = createSourceNodeAt(Root(),src,"a.foo")

        module = getTypeOf(root,"a.foo")
        refs = [x for x in findReferences(module.getSourceNode(),2,8)
                if x.confidence == 100]
        self.assertEqual(refs[0].filename,
                         os.path.abspath(os.path.join("a","foo.py")))
        self.assertEqual(refs[0].modulefqn,"a.foo")
        self.assertEqual(refs[0].lineno,3)
        self.assertEqual(refs[0].colno,11)
        self.assertEqual(refs[0].colend,20)


class FindReferencesToClass(unittest.TestCase):
    def test_returnsEmptyListIfNoReferences(self):
        src = trimLines("""
        class MyClass:
            pass
        a = TheClass()
        """)
        root = createSourceNodeAt(Root(),src,"mymodule")
        module = getTypeOf(root,"mymodule")
        refs = [x for x in findReferences(module.getSourceNode(),1,6)]
        assert refs == []


    def test_findsSimpleReferenceInSameModule(self):
        src = trimLines("""
        class TheClass:
            pass
        a = TheClass()
        """)
        root = createSourceNodeAt(Root(),src,"mymodule")
        module = getTypeOf(root,"mymodule")
        refs = [x for x in findReferences(module.getSourceNode(),1,6)]
        self.assertEqual(refs[0].filename,os.path.abspath("mymodule.py"))
        self.assertEqual(refs[0].modulefqn,"mymodule")
        self.assertEqual(refs[0].lineno,3)
        self.assertEqual(refs[0].colno,4)
        self.assertEqual(refs[0].confidence,100)

    def test_doesntBarfOnSingleLineSourceWithInlineClass(self):
        src=trimLines("""
        from b.bah import TheClass
        a = TheClass()
        """)
        root = createSourceNodeAt(Root(),src,"a.foo")
        root = createSourceNodeAt(root, "class TheClass: pass", "a.b.bah")
        module = getTypeOf(root,"a.b.bah")
        refs = [x for x in findReferences(module.getSourceNode(),1,6)]
        assert refs != []

    def test_findsReferenceToClassImportedInSameClassScope(self):
        src=trimLines("""
        class AnotherClass:
            from b.bah import TheClass
            TheClass.baz = 0
        """)
        root = createSourceNodeAt(Root(),src,"a.foo")
        root = createSourceNodeAt(root, testClass, "a.b.bah")
        module = getTypeOf(root,"a.b.bah")
        refs = [x for x in findReferences(module.getSourceNode(),1,6)]

        self.assertEqual(refs[0].filename,
                         os.path.abspath(os.path.join("a","foo.py")))
        self.assertEqual(refs[0].modulefqn,"a.foo")
        self.assertEqual(refs[0].lineno,2)
        self.assertEqual(refs[0].colno,22)

        self.assertEqual(refs[1].filename,
                         os.path.abspath(os.path.join("a","foo.py")))
        self.assertEqual(refs[1].modulefqn,"a.foo")
        self.assertEqual(refs[1].lineno,3)
        self.assertEqual(refs[1].colno,4)

    def testFindsClassReferenceWhenScopeIsSameNameAsClass(self):
        src = trimLines("""
        class TheClass:
            class TheClass:
                pass
        a = TheClass.TheClass()
        """)
        root = createSourceNodeAt(Root(),src,"a.foo")
        module = getTypeOf(root,"a.foo")
        refs = [x for x in findReferences(module.getSourceNode(),2,10)]
        self.assertEqual(refs[0].filename,
                         os.path.abspath(os.path.join("a","foo.py")))
        self.assertEqual(refs[0].lineno,4)
        self.assertEqual(refs[0].colno,13)
        self.assertEqual(refs[0].confidence,100)

    def testFindsClassReferenceWhenChildIsSameNameAsClass(self):
        src = trimLines("""
        class TheClass:
            class TheClass:
                pass
        a = TheClass.TheClass()
        """)
        root = createSourceNodeAt(Root(),src,"a.foo")
        module = getTypeOf(root,"a.foo")
        refs = [x for x in findReferences(module.getSourceNode(),1,6)]
        self.assertEqual(refs[0].filename,
                         os.path.abspath(os.path.join("a","foo.py")))
        self.assertEqual(refs[0].lineno,4)
        self.assertEqual(refs[0].colno,4)
        self.assertEqual(refs[0].confidence,100)

    '''
    def test_findsReferencesInModuleWhichImportsClass(self):
        src = trimLines("""
        import b.bah.TheClass
        def foo():
            a = b.bah.TheClass()
            a.theMethod()
        """)
        root = Root()
        root = createSourceNodeAt(root, src, "a.foo")
        root = createSourceNodeAt(root, testClass, "a.b.bah")
        module = getTypeOf(root,"a.b.bah")
        refs = [x for x in findReferences(module.getSourceNode(),1,6)]
        self.assertEqual(refs[0].filename,os.path.abspath(os.path.join("a","foo.py")))
        self.assertEqual(refs[0].modulefqn,"a.foo")
        self.assertEqual(refs[0].lineno,1)
        self.assertEqual(refs[0].colno,13)
        self.assertEqual(refs[0].confidence,100)

        self.assertEqual(refs[1].filename,os.path.abspath(os.path.join("a","foo.py")))
        self.assertEqual(refs[1].modulefqn,"a.foo")
        self.assertEqual(refs[1].lineno,3)
        self.assertEqual(refs[1].colno,14)
        self.assertEqual(refs[0].confidence,100)
    '''

class TestFindReferencesIncludingDefn(unittest.TestCase):
    def test_findsMethodDecl(self):
        src=trimLines("""
        class TheClass:
            def theMethod(self):
                pass
        """)
        root = createSourceNodeAt(Root(),src,"a.foo")
        module = getTypeOf(root,"a.foo")
        refs = [x for x in findReferencesIncludingDefn(module.getSourceNode(),2,8)]
        self.assertEqual(refs[0].filename,
                         os.path.abspath(os.path.join("a","foo.py")))
        self.assertEqual(refs[0].lineno,2)
        self.assertEqual(refs[0].colno,8)
        self.assertEqual(refs[0].confidence,100)


testClass = trimLines("""
class TheClass:
    def theMethod(self):
        pass
    def differentMethod(self):
        pass

class DifferentClass:
    def theMethod(self):
        pass
""")



if __name__ == "__main__":
    unittest.main()
